#!/usr/bin/env python3

import re
import os
import sys
import getopt
import shutil
import json
import ast

import lib.json2csv as jsontoolkit

#############################
#     GLOBAL  VARIABLES     #
#############################

# Build OS portable paths for the project and input/output dirs
filepath = os.path.realpath(__file__)
projectpath = os.path.dirname(filepath)
output_dir = os.path.join(projectpath, "output")

# OS portable paths for default filters
default_exclude_filter = os.path.join(projectpath, "data", "filters",
                                      "default_exclude_filter.txt")
default_include_filter = os.path.join(projectpath, "data", "filters",
                                      "default_include_filter.txt")

# Valid execution modes
execution_modes = ["split", "flatten", "flat2csv", "jsoncollection2csv",
                   "singlejson2csv"]

# Default execution mode
mode = "jsoncollection2csv"

# Verbosity and silence flags
verbose = False
silent = False

# Array compression mode (for CSV conversion)
array_compression = False
compression_depth = 0
# Remove dollar terminations inserted by Mongo
remove_dollars = False
# Filtering flag
filtering = False
# Keep intermediate files flag
keep_files = False


#############################
#    AUXILIAR FUNCTIONS     #
#############################
def bad_invocation():
    print("Unacceptable invocation. To know more about the syntax:")
    print("./main.py -h")
    sys.exit(2)


def print_help():
    """
    Basic help about usage of the program and accepted syntax that will
    be printed when the option -h or --help is received
    """

    print("Usage:\n\t./main.py [arguments] file")

    print("\nArguments:")
    print("\n\t-h | --help\n\t\tPrint help about syntax and usage")
    print("\n\t-m <mode> | --mode=<mode>\n\t\tSelect an execution " +
          "mode")
    print("\n\t-o <dir> | --output=<dir>\n\t\tSelect an output " +
          "directory")
    print("\n\t-c <depth> | --array-compression <depth>\n" +
          "\t\tCompress the array fields in a single list in the CSV. " +
          "Follow with\n\t\tan integer arg indicating the number of " +
          "nested arrays to compress.\n\t\tUse an arbitrarily large " +
          "number to indicate full compression")
    print("\n\t-r | --remove-dollars\n\t\tRemove the dollar-" +
          "headed terminations")
    print("\n\t-u | --upv-compatible\n\t\tEquivalent to " +
          "remove-dollars + array-compression + default include filter")
    print("\n\t-v | --verbose\n\t\tVerbose mode")
    print("\n\t-s | --silent\n\t\tSilent mode. If this flag is " +
          "present the verbose flag will be muted and\n\t\teven " +
          "the most critical messages will be silenced.")
    print("\n\t-f | --default-include-filter\n\t\tUse the " +
          "default include filter")
    print("\n\t-F | --default-exclude-filter\n\t\tUse the " +
          "default exclude filter")
    print("\n\t-k | --keep-files\n\t\tKeep the intermediate " +
          "files (split and flat) when doing full cycle JSON to " +
          "CSV conversion")
    print("\nModes:")
    print("\tsplit\t\t\t\tSplit a collection (single file) of JSON" +
          " objects in individual files")
    print("\tflatten\t\t\t\tFlatten a collection of JSON files")
    print("\tflat2csv\t\t\tConvert from flat JSON to CSV")
    print("\tjsoncollection2csv\t\tConvert a JSON collection (single file " +
          "containing multiple JSON items) to CSV")
    print("\tsinglejson2csv\t\t\tConvert a single JSON object " +
          "to CSV (doesn't generate a directory tree)")
    sys.exit()


def change_output_dir(output):
    """
    Change the final directory where the output will be saved if the
    user specified so. However, intermediate results will still be
    stored in the traditional dirs
    """

    if os.path.isdir(output):
        if verbose:
            print("The resulting files will be stored at",
                  os.path.abspath(output))
        return output
    else:
        if not silent:
            print("The specified output is not an existing system " +
                  "directory. Check the spelling or create the " +
                  "folder if it doesn't exist yet.")
        sys.exit(2)


def change_execution_mode(mode):
    """
    If specified by the user, change execution mode. Else, default is
    jsoncollection2csv
    """

    if mode in execution_modes:
        if verbose:
            print("Mode of execution set to", mode)
        return mode
    else:
        if not silent:
            print("The selected mode of execution does not exist. Stopping")
        sys.exit(2)


def load_filter(filter_path):
    """
    Load the file specified in the argument and compile its regexes.
    Return the list of compiled regexes that form the ready-to-use
    filter
    """

    temporary_filter = []
    with open(filter_path, 'r') as dictionary_file:
        filter_dict = ast.literal_eval(dictionary_file.read())

    for item in filter_dict:
        temporary_filter.append(re.compile(item))

    return temporary_filter


def garbage_collector(split_input, flat_input):
    if split_input is not None:
        if os.path.isdir(split_input):
            shutil.rmtree(split_input)
        elif os.path.isfile(split_input):
            os.remove(split_input)
    if flat_input is not None:
        if os.path.isdir(flat_input):
            shutil.rmtree(flat_input)
        elif os.path.isfile(flat_input):
            os.remove(flat_input)


def safe_copyfile(copiable_item, copy_path):
    """
    Given a path where a file is going to be copied (including the filename),
    add an underscore symbol followed by a number. To chose the number iterate
    numbers from 0 to infinite until it finds a free name. For example, if
    foo.txt exists in the destination path and so do foo_1.txt and foo_2.txt,
    this function should return foo_3.txt
    """

    i = 1

    if os.path.isfile(copy_path):
        new_copy_path = copy_path
        copy_path = None

        # Matches patterns such as: '/home/user/foo.json' or '/bin/foo.tar.gz'
        # Grouping '/bin/foo' in group 1 and '.tar.gz' in group 2
        matchObj =\
            re.search(r'(.*[\\\/][^\\\/\.]+)(\.[^\\\/]+)$',
                      new_copy_path)
        new_copy_path = matchObj.group(1) + "_" + str(i) +\
            matchObj.group(2)

        while os.path.isfile(new_copy_path):
            # Same match than before but including a parethesized number before
            # the extension. For example, '/bin/foo(4).tar.gz
            matchObj =\
                re.search(r'(.*[\\\/][^\\\/\.]+)(_\d+)(\.[^\\\/]+)$',
                          new_copy_path)
            new_copy_path = matchObj.group(1) + "_" + str(i) +\
                matchObj.group(3)
            i += 1

    shutil.copyfile(copiable_item, copy_path or new_copy_path)


def create_collection_dir(input_file, final_output_dir):
    """
    Given the input file (either a file of JSONs, a single JSON, a collection
    of JSONs or a collection of flat JSONs) and an output directory, create a
    new folder in that directory with the name of the collection. This will be
    the top of the new tree of files
    """

    if os.path.isfile(input_file) and mode in ("split", "jsoncollection2csv"):

        # Use regex to get only the basename of the file and build an output
        # path. If there is no match, file isn't a JSON. Throw error
        matchObj = re.search(r'.*[/\\](.*)\.json$', input_file)

        if not matchObj:
            if not silent:
                print_not_json()
            sys.exit(1)

        basename = matchObj.group(1)
        collection_path = os.path.join(final_output_dir, basename)

        try:
            os.mkdir(collection_path)
        except OSError:
            pass

    elif os.path.isfile(input_file) and mode not in ("split"):
        collection_path = final_output_dir

    elif os.path.isdir(input_file):
        collection_name = decide_collection_name(input_file)
        collection_path = os.path.join(final_output_dir, collection_name)

        try:
            os.mkdir(collection_path)
        except OSError:
            pass

    else:
        print("Path " + input_file +
              " does not point to a file nor a directory")
        sys.exit(2)

    return collection_path


def decide_collection_name(input_dir):
    """
    Decide in a smart way what is going to be the name of the new collection
    """

    original_dir = os.path.normpath(input_dir)
    matchObj = re.search(r'^.*[/\\]([^/\\]+)[/\\]([^/\\]+)$', original_dir)
    if matchObj.group(2) in ("Split", "Flat"):
        collection_name = matchObj.group(1)
    else:
        collection_name = matchObj.group(2)

    return collection_name


def create_subdir(output_dir, dir_name=""):
    """
    This is used to create the three subfolders present in each collection:
    "Split", "Flat" and "CSV".
    """

    if dir_name:
        output_dir = os.path.join(output_dir, dir_name)

    # Create a new directory for the collection unless it already exists
    try:
        os.mkdir(output_dir)
    except OSError:
        handle_existing_collection(output_dir)

    return output_dir


##############################
#     PRINTED MESSAGES       #
##############################
def splitting_collection_msg(input_, output):
    print("Splitting the collection "
          + os.path.realpath(input_) +
          "\nAnd storing the results in:\n"
          + os.path.realpath(output))


def flattening_collection_msg(input_, output):
    print("\nFlattening the collection:"
          + os.path.realpath(input_) +
          "\nAnd storing the results in:\n"
          + os.path.realpath(output))


def converting2csv_msg(input_, output):
    print("\nConverting the following to CSV format:"
          + os.path.realpath(input_) +
          "\nWith compression mode = " + str(array_compression) +
          "\nAnd remove dollars mode = " + str(remove_dollars) +
          "\nAnd storing the results in:\n"
          + os.path.realpath(output))


def print_invalid_json_syntax():
    print("Something seems to be wrong with your JSON " +
          "file. Please, check its structure and syntax." +
          "\nAborting...")


def print_not_json():
    print("Please, check that the input file you specify " +
          "has a .json extension.\nAborting...")


def print_not_flat():
    print("The file you are trying to convert to CSV seems"
          + " to contain nested dictionaries.\n" +
          "Try flattening it first.")


##############################
#      ERROR HANDLERS        #
##############################
def handle_existing_collection(output_path):
    """
    Use this to handle the exception thrown when the system detects an
    already existent split collection with the same path that the one we
    are going to create. Given that case, prompt the user for a
    decission on overwriting
    """

    if not silent:
        print("\nThe directory you selected as output for the new " +
              "collection does already contain one with the same name.")
    if input("Do you really want to overwrite it and lose the previous " +
             "files?(Y/N) > ").lower().startswith('y'):

        shutil.rmtree(output_path)
        return create_subdir(output_path)
    else:
        if not silent:
            print("Aborting...")
        sys.exit(1)


def handle_existing_split_collection(input_, output, collection_path):
    """
    Use this to handle the exception thrown when the system detects an
    already existent split collection with the same path that the one we
    are going to create. Given that case, prompt the user for a
    decission on overwriting
    """

    if not silent:
        print("\nThe directory you selected as output for the new " +
              "split collection does already contain one with the same name.")
    if input("Do you really want to overwrite it and lose the previous " +
             "files?(Y/N) > ").lower().startswith('y'):

        shutil.rmtree(collection_path)
        return jsontoolkit.split_collection(input_, output)
    else:
        if not silent:
            print("Aborting...")
        sys.exit(1)


def handle_existing_flat_collection(input_, output, collection_path):
    """
    Use this to handle the exception thrown when the system detects an
    already existent flat collection with the same path that the one we
    are going to create. Given that case, prompt the user for a
    decission on overwriting
    """

    if not silent:
        print("\nThe directory you selected as output for the new " +
              "flat collection does already contain one with the same name.")
    if input("Do you really want to overwrite " +
             "it and lose the previous files? (Y/N) > ").lower().\
            startswith('y'):

        shutil.rmtree(collection_path)
        return jsontoolkit.flatten_dir_files(input_, output)
    else:
        if not silent:
            print("Aborting...")
        sys.exit(1)


def handle_existing_csv_collection(input_, output,
                                   include_filter, exclude_filter,
                                   collection_path):
    """
    Use this to handle the exception thrown when the system detects an
    already existent CSV collection with the same path that the one we
    are going to create. Given that case, prompt the user for a
    decission on overwriting
    """

    if not silent:
        print("\nThe directory you selected as output for the new " +
              "CSV collection does already contain one with the " +
              "same name.")
    if input("Do you really want to overwrite " +
             "it and lose the previous files? (Y/N) > ").lower().\
            startswith('y'):

        shutil.rmtree(collection_path)
        return jsontoolkit.convert_dir_json2csv(input_, output,
                                                array_compression,
                                                compression_depth,
                                                remove_dollars,
                                                include_filter,
                                                exclude_filter)
    else:
        if not silent:
            print("Aborting...")
        sys.exit(1)


def handle_existing_csv_file(input_, output, include_filter, exclude_filter,
                             previous_file):

    if not silent:
        print("\nThe file that was going to be created does already exist")
    if input("Do you really want to overwrite " +
             "it and lose the previous file? (Y/N) > ").lower().\
            startswith('y'):
        os.remove(previous_file)
        return singlejson2csv_mode(input_, output,
                                   include_filter, exclude_filter)
    else:
        if not silent:
            print("Aborting...")
        sys.exit(1)


##############################
#      EXECUTION MODES       #
##############################
def split_mode(input_, output):
    """
    If the input is a file, make a call to the JSON toolkit library to
    split the file in single JSONs and store the results in the
    specified output dir
    """

    # For splitting only files (JSON collections) are admitted
    if not os.path.isfile(input_):
        if not silent:
            print("You need to specify an input file to split")
        sys.exit(2)

    else:
        if verbose:
            print("Files from the collection:\n"
                  + os.path.realpath(input_) +
                  "\nWill be split into single JSONs and stored in:\n"
                  + os.path.realpath(output))

        try:
            # Splitting algorithm call
            output = create_subdir(output, "Split")
            return jsontoolkit.split_collection(input_, output)

        # Handle possible exceptions
        except jsontoolkit.WrongFiletypeException:
            if not silent:
                print_not_json()
            shutil.rmtree(output)
            sys.exit(1)
        except SyntaxError:
            if not silent:
                print_invalid_json_syntax()
            shutil.rmtree(output)
            sys.exit(1)
        except OSError as e:
            return handle_existing_split_collection(input_, output, str(e))


def flatten_mode(input_, output):
    """
    If the input is a file or directory, make a call to the JSON toolkit
    library to flatten the file or collection and store the results in
    the specified output dir
    """

    # If path is not valid throw error
    if not (os.path.isfile(input_) or os.path.isdir(input_)):
        if not silent:
            print("You need to specify a valid path to a file or dir")
        sys.exit(2)

    # For flattening mode files are interpreted as splitted JSONs
    else:
        if verbose:
            print("Performing a flattening operation on:\n"
                  + os.path.realpath(input_) +
                  "\nAnd storing the results in:\n"
                  + os.path.realpath(output))

        # Flatten single file
        if os.path.isfile(input_):
            try:
                return jsontoolkit.flatten_file(input_, output)

            except jsontoolkit.WrongFiletypeException:
                if not silent:
                    print_not_json()
                sys.exit(1)
            except json.decoder.JSONDecodeError:
                if not silent:
                    print_invalid_json_syntax()
                sys.exit(1)
            except OSError as e:
                return handle_existing_flat_collection(input_, output, str(e))

        # Flatten all files in a directory
        elif os.path.isdir(input_):
            try:
                output = create_subdir(output, "Flat")
                return jsontoolkit.flatten_dir_files(input_, output)

            except jsontoolkit.WrongFiletypeException:
                if not silent:
                    print_not_json()
                shutil.rmtree(output)
                sys.exit(1)
            except json.decoder.JSONDecodeError:
                if not silent:
                    print_invalid_json_syntax()
                shutil.rmtree(output)
                sys.exit(1)
            except OSError as e:
                return handle_existing_flat_collection(input_, output, str(e))


def flat2csv_mode(input_, output, include_filter=[], exclude_filter=[]):
    """
    If the input is a file or directory, make a call to the JSON toolkit
    library to convert the file or collection from JSON to CSV format
    and store the results in the specified output dir
    """

    # If path is not valid throw error
    if not (os.path.isfile(input_) or os.path.isdir(input_)):
        if not silent:
            print("You need to specify a valid path to a file or dir")
        sys.exit(2)

    else:
        if verbose:
            print("Performing a JSON-CSV conversion on:\n"
                  + os.path.realpath(input_) +
                  "\nAnd storing the results in:\n"
                  + os.path.realpath(output))

        try:
            # Convert to CSV a single file (should be flat JSON)
            if os.path.isfile(input_):
                return jsontoolkit.convert_json2csv(
                    input_, output, array_compression, compression_depth,
                    remove_dollars, include_filter, exclude_filter)

            # Convert to CSV all files in a dir
            elif os.path.isdir(input_):
                output = create_subdir(output, "CSV")
                return jsontoolkit.convert_dir_json2csv(
                    input_, output, array_compression, compression_depth,
                    remove_dollars,
                    include_filter, exclude_filter)

        except jsontoolkit.WrongFiletypeException as e:
            if not silent:
                print_not_json()

            if os.path.isdir(str(e)):
                matchObj =\
                    re.search(r'(.*)/\d{4}/\d{2}/\d{2}/?$', str(e))
                if matchObj:
                    shutil.rmtree(matchObj.group(1))
                else:
                    shutil.rmtree(str(e))

            sys.exit(1)

        # This is susceptible of being changed to another way of
        # managing the exception (perhaps redirect to flattener or ask
        # the user for a decision)
        except jsontoolkit.NotFlattenedException as e:
            if not silent:
                print_not_flat()

            if os.path.isdir(str(e)):
                matchObj =\
                    re.search(r'(.*)/\d{4}/\d{2}/\d{2}/?$', str(e))
                if matchObj:
                    shutil.rmtree(matchObj.group(1))
                else:
                    shutil.rmtree(str(e))
            sys.exit(1)

        except OSError as e:
            handle_existing_csv_collection(input_, output, include_filter,
                                           exclude_filter, str(e))


def jsoncollection2csv_mode(input_, output,
                            include_filter=[], exclude_filter=[]):
    """
    If the input is a file or directory, make a call to the JSON toolkit
    library to split the collection (if needed), flatten the file(s) and
    convert them to CSV format to finally store the results in the
    specified output dir
    """

    # If path is not valid throw error
    if not (os.path.isfile(input_) or os.path.isdir(input_)):
        if not silent:
            print("You need to specify a valid path to a file or dir")
        sys.exit(2)

    else:
        split_collection = None
        # If input_ points to a file, split
        if os.path.isfile(input_):
            # Splitting file
            split_output = create_subdir(output, "Split")
            if verbose:
                splitting_collection_msg(input_, split_output)
            split_collection =\
                jsontoolkit.split_collection(input_, split_output)

        # Flattening file
        flat_output = create_subdir(output, "Flat")
        if verbose:
            flattening_collection_msg(split_collection, flat_output)
        flat_collection = jsontoolkit.flatten_dir_files(
            split_collection or input_, flat_output)

        # Finally, convert to CSV and store where needed
        csv_output = create_subdir(output, "CSV")
        if verbose:
            converting2csv_msg(flat_collection, csv_output)
        csv_collection =\
            jsontoolkit.convert_dir_json2csv(flat_collection, csv_output,
                                             array_compression,
                                             compression_depth, remove_dollars,
                                             include_filter, exclude_filter)

        if keep_files is False:
            garbage_collector(split_collection, flat_collection)

        return csv_collection


def singlejson2csv_mode(input_, output, include_filter=[], exclude_filter=[]):
    """
    If the input is a single json object, make a call to the JSON toolkit
    library to flatten the file and convert it to CSV format to finally store
    the results in the specified output dir
    """

    # If path is not valid throw error
    if not os.path.isfile(input_):
        if not silent:
            print("You need to specify a valid path to a file or dir")
        sys.exit(2)

    else:
        # Make a temporal dir to store the flat file
        if verbose:
            converting2csv_msg(input_, output)
        tmp_dir = os.path.join(output, "tmp")
        try:
            os.mkdir(tmp_dir)
            flat_json = jsontoolkit.flatten_file(input_, tmp_dir)
        except OSError as e:
            if not silent:
                print("The following directory is a system folder and " +
                      "should not be created by the user" + str(e))
            sys.exit(2)
        except jsontoolkit.WrongFiletypeException:
            if not silent:
                print_not_json
            shutil.rmtree(tmp_dir)
            sys.exit(1)
        except json.decoder.JSONDecodeError:
            if not silent:
                print_invalid_json_syntax
            shutil.rmtree(tmp_dir)
            sys.exit(1)

        try:
            csv_file =\
                jsontoolkit.convert_json2csv(flat_json, output,
                                             array_compression,
                                             compression_depth, remove_dollars,
                                             include_filter, exclude_filter)
            # If the user wants the flat file, move it to the output dir
            if keep_files:
                copy_path = os.path.join(output, os.path.basename(flat_json))
                safe_copyfile(flat_json, copy_path)
            shutil.rmtree(tmp_dir)
            garbage_collector(None, flat_json)

            return csv_file

        except OSError as e:
            shutil.rmtree(tmp_dir)
            return handle_existing_csv_file(input_, output, include_filter,
                                            exclude_filter, str(e))

#############################
#       MAIN SCRIPT         #
#############################


def main():

    # VARIABLE DECLARATION
    final_output_dir = output_dir
    include_filter = False
    exclude_filter = False
    regex_include_filter = []
    regex_exclude_filter = []

    global verbose
    global silent
    global array_compression
    global compression_depth
    global remove_dollars
    global filtering
    global keep_files
    global mode

    # Program needs some args
    if len(sys.argv) <= 1:
        bad_invocation()

    # Select the input file or dir (last arg)
    input_file = sys.argv.pop()
    if input_file in ("-h", "--help"):
        print_help()

    # Option & argument parsing
    try:
        opts, args = getopt.getopt(sys.argv[1:], "kfFrusvhc:m:o:e:i:",
                                   [
                                    "keep-files",
                                    "default-include-filter",
                                    "default-exclude-filter",
                                    "remove-dollars", "upv-compatible",
                                    "silent", "verbose", "help",
                                    "array-compression=", "mode=", "output=",
                                    "exclude", "include"
                                   ])

    except getopt.GetoptError:
        bad_invocation()

    for opt, arg in opts:
        if opt in ("-h", "--help"):
            print_help()
        else:
            if opt in ('-m', "--mode"):
                mode = change_execution_mode(arg)
            if opt in ('-o', "--output"):
                final_output_dir = change_output_dir(arg)
            if opt in ('-v', "--verbose"):
                verbose = True
            if opt in ('-s', "--silent"):
                silent = True
            if opt in ('-c', "--array-compression"):
                array_compression = True
                try:
                    compression_depth = int(arg)
                except ValueError:
                    print("Array compression option should have an integer " +
                          "value indicating depth. Invoke 'main.py -h' for " +
                          "further details")
                    sys.exit(2)

            if opt in ('-r', "--remove-dollars"):
                remove_dollars = True
            if opt in ('-u', "--upv-compatible"):
                array_compression = True
                compression_depth = 1
                remove_dollars = True
                if not include_filter:
                    include_filter = default_include_filter
            if opt in ('-e', "--exclude"):
                exclude_filter = arg
            if opt in ('-i', "--include"):
                include_filter = arg
            if opt in ('-f', "--default-include-filter"):
                if not include_filter:
                    include_filter = default_include_filter
            if opt in ('-F', "--default-exclude-filter"):
                if not exclude_filter:
                    exclude_filter = default_exclude_filter
            if opt in ('-k', "--keep-files"):
                keep_files = True

    # Verbosity and silence are incompatible. Silence prevails
    if silent:
        verbose = False

    # Load filters, if any
    if include_filter:
        regex_include_filter = load_filter(include_filter)
    elif exclude_filter:
        regex_exclude_filter = load_filter(exclude_filter)

    # Create a dir to store the collection
    final_output_dir = create_collection_dir(input_file, final_output_dir)

    # Modes of execution
    if mode == "split":
        split_mode(input_file, final_output_dir)

    elif mode == "flatten":
        flatten_mode(input_file, final_output_dir)

    elif mode == "flat2csv":
        flat2csv_mode(input_file, final_output_dir,
                      regex_include_filter, regex_exclude_filter)

    elif mode == "singlejson2csv":
        singlejson2csv_mode(input_file, final_output_dir,
                            regex_include_filter, regex_exclude_filter)

    elif mode == "jsoncollection2csv":
        jsoncollection2csv_mode(input_file, final_output_dir,
                                regex_include_filter, regex_exclude_filter)


#############################
#       MAIN EXECUTION      #
#############################
if __name__ == "__main__":
    main()
