#!/usr/bin/env python3

import unittest
import os
import shutil
import re
import filecmp
import csv
import time

from unittest import mock
from context import main


class MainTestCase(unittest.TestCase):
    """Tests for the main execution script"""

    def setUp(self):
        """
        Set up some variables
        """

        self.this_file = os.path.realpath(__file__)
        self.this_dir = os.path.dirname(self.this_file)
        self.input_dir = os.path.join(self.this_dir, "input")

        # Run the tests silently to avoid filling screen with text
        main.silent = True

    def test_changeOutputDir_toRealDir_OK(self):
        """Test that change_output_dir returns the new dir"""

        output_dir = main.change_output_dir(self.this_dir)
        self.assertEqual(output_dir, self.this_dir)

    def test_changeOutputDir_toFile_fails(self):
        """
        Test that change_output_dir makes sys.exit(2) when someone
        passes a file instead of a dir
        """
        with self.assertRaises(SystemExit) as context_manager:
            main.change_output_dir(self.this_file)
        self.assertEqual(context_manager.exception.code, 2)

    def test_changeOutputDir_toJunk_fails(self):
        """
        Test that change_output_dir makes sys.exit(2) when someone
        passes random things
        """
        with self.assertRaises(SystemExit) as context_manager:
            main.change_output_dir("waza wasa")
        self.assertEqual(context_manager.exception.code, 2)

    def test_changeExecutionMode_toFlatten_OK(self):
        """
        Test that change_execution_mode to some of the existent
        execution modes works well
        """

        self.assertEqual(main.change_execution_mode("flatten"), "flatten")

    def test_changeExecutionMode_toRamen_fails(self):
        """
        Test that change_execution_mode to something else fails
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.change_execution_mode("Ramen")
        self.assertEqual(context_manager.exception.code, 2)


class SplitTestCase(unittest.TestCase):

    def setUp(self):
        """
        Set up some variables
        """

        self.this_file = os.path.realpath(__file__)
        self.this_dir = os.path.dirname(self.this_file)
        self.input_dir = os.path.join(self.this_dir, "input")
        self.json_collection = os.path.join(self.input_dir,
                                            "ChicagoGeoTime_Brexit.json")
        self.ten_tweets_file = os.path.join(self.input_dir, "ten_tweets.json")
        self.two_tweets_file = os.path.join(self.input_dir, "two_tweets.json")
        self.single_tweet_file = os.path.join(self.input_dir,
                                              "single_tweet.json")
        self.not_json_file = os.path.join(self.input_dir, "text_file.txt")
        self.bad_json_file = os.path.join(self.input_dir,
                                          "corrupt_json_file.json")
        self.not_tweet_collection = os.path.join(self.input_dir,
                                                 "not_tweet_collection.json")
        main.silent = True

    def test_splitMode_jsonCollection_OK(self):
        """
        Test that the activation of split mode works fine with a more or
        less large JSON file. Then remove the resulting collection dir.
        """
        split_col = main.split_mode(self.json_collection, self.this_dir)
        shutil.rmtree(split_col)

    def test_splitMode_tenTweetsCol_OK(self):
        """
        Test that the splitting of a mini-collection of ten tweets works
        fine and results in a new dir with ten splitted JSONs.
        """

        split_col = main.split_mode(self.ten_tweets_file, self.this_dir)
        self.assertEqual(len(os.listdir(split_col)), 10)
        shutil.rmtree(split_col)

    def test_splitMode_singleTweet_OK(self):
        """
        Even though the split method was thought for more or less large
        collections, it should work well as well with a single tweet
        (which simply copies the tweet)
        """

        single_tweet = main.split_mode(self.single_tweet_file, self.this_dir)
        self.assertEqual(len(os.listdir(single_tweet)), 1)
        self.assertEqual(os.listdir(single_tweet)[0], "0.json")
        shutil.rmtree(single_tweet)

    def test_splitMode_inputNotJson_fails(self):
        """
        If the input is something else (not a .json file) the program
        should abort and throw error
        """

        with self.assertRaises(SystemExit) as context_manager:
            output_dir =\
                main.create_collection_dir(self.not_json_file, self.this_dir)
            main.split_mode(self.not_json_file, output_dir)
        self.assertEqual(context_manager.exception.code, 1)

    def test_splitMode_inputCorruptJson_fails(self):
        """
        If the input is a bad JSON (in this case we only check that the
        number of braces is good), the program should fail appropriately and no
        collection dir should be created
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.split_mode(self.bad_json_file, self.this_dir)
        self.assertEqual(context_manager.exception.code, 1)
        self.assertFalse(os.path.isdir(os.path.join(self.this_dir, "Split")))

    def test_splitMode_collectionAlreadyExistsOverwrite_OK(self):
        """
        Try to split a collection twice: the first time it should create
        the output dir normally, and the second one it will find the
        previous dir. Mock up a 'y' response to the overwrite prompt and
        remove the new directory.
        Care!! If a dir for that collection previously existed this
        could fail
        """

        with mock.patch('main.input', side_effect=['y']):

            split_col = main.split_mode(self.two_tweets_file, self.this_dir)
            split_col = main.split_mode(self.two_tweets_file, self.this_dir)

        if 'split_col' in locals():
            shutil.rmtree(split_col)

    def test_splitMode_collectionAlreadyExistsAbort_OK(self):
        """
        Try to split a collection twice: the first time it creates a new
        collection and the second one it answers "n" to the overwriting
        question. Program should abort and exit with code 1
        """

        with self.assertRaises(SystemExit) as context_manager:
            with mock.patch('main.input', side_effect=['n']):

                splt_col = main.split_mode(self.two_tweets_file, self.this_dir)
                splt_col = main.split_mode(self.two_tweets_file, self.this_dir)

        if 'splt_col' in locals():
            shutil.rmtree(splt_col)

        self.assertEqual(context_manager.exception.code, 1)

    def test_splitMode_notTweetJsonCollection_OK(self):
        """
        Splitter should be good for any type of JSON files, not just
        tweet objects. This test checks that everything is alright with
        other data not taken from Twitter, and when split creates a
        collection with JSON files named with numbers
        """

        split_col =\
            main.split_mode(self.not_tweet_collection, self.this_dir)
        for filename in os.listdir(split_col):
            self.assertRegex(filename, r'\d+\.json$')
        shutil.rmtree(split_col)


class FlattenTestCase(unittest.TestCase):

    def setUp(self):
        """Set up some variables and files"""

        self.this_file = os.path.realpath(__file__)
        self.this_dir = os.path.dirname(self.this_file)
        self.input_dir = os.path.join(self.this_dir, "input")

        # Same tweet before and after flattening
        # These files have been checked for correctness
        self.json_file = os.path.join(self.input_dir, "single_tweet.json")
        self.json_flat_file = os.path.join(self.input_dir,
                                           "single_tweet_flat.json")
        self.not_json_file = os.path.join(self.input_dir, "text_file.txt")
        self.bad_json_file = os.path.join(self.input_dir,
                                          "corrupt_json_file.json")
        self.json_collection = os.path.join(self.input_dir, "ten_tweets")
        self.not_tweet_json_file = os.path.join(self.input_dir,
                                                "not_tweet_file.json")
        self.not_tweet_json_flat_file =\
            os.path.join(self.input_dir, "not_tweet_file_flat.json")
        self.not_tweet_json_collection =\
            os.path.join(self.input_dir, "not_tweet_collection")
        self.not_json_dir = os.path.join(self.input_dir, "not_json_dir")

        main.silent = True

    def test_flattenMode_singleFile_namesOK(self):
        """
        Trying to flatten a single JSON file should perform correctly
        and return a flattened JSON file. Check that the name equals the
        name of the previous file with a '_flat' added before the file
        extension
        """

        json_file_name =\
            os.path.basename(os.path.normpath(self.json_file))
        matchObj = re.search(r'(.*)(\.json)', json_file_name)

        flat_file = main.flatten_mode(self.json_file, self.this_dir)
        flat_file_name = os.path.basename(os.path.normpath(flat_file))

        self.assertEqual(flat_file_name,
                         matchObj.group(1)+"_flat"+matchObj.group(2))

        os.remove(flat_file)

    def test_flattenMode_singleFile_filesEqualOK(self):
        """
        Using a previously flattened JSON which has already been checked
        for correctness, check that the file resulting from the
        flattening process is equal to that file
        """

        flat_file = main.flatten_mode(self.json_file, self.this_dir)
        are_equal = filecmp.cmp(self.json_flat_file, flat_file)
        self.assertTrue(are_equal)

        os.remove(flat_file)

    def test_flattenMode_collection_treeOK(self):
        """
        Given a collection of split JSONs, call the flatten function and
        perform some checks over the resulting directory and file
        system. The new tree should have the following structure: years
        > months > days > flattened JSON files. We check this using
        regexes
        """

        flat_collection =\
            main.flatten_mode(self.json_collection, self.this_dir)

        for year_dir in os.listdir(flat_collection):
            year_dir_path = os.path.join(flat_collection, year_dir)

            for month_dir in os.listdir(year_dir_path):
                month_dir_path = os.path.join(year_dir_path, month_dir)

                for day_dir in os.listdir(month_dir_path):
                    day_dir_path = os.path.join(month_dir_path, day_dir)

                    for json_file in os.listdir(day_dir_path):

                        self.assertRegex(day_dir, r'\d{2}')
                        self.assertRegex(month_dir, r'\d{2}')
                        self.assertRegex(year_dir, r'\d{4}')
                        self.assertRegex(json_file, r'.*\.json')

        shutil.rmtree(flat_collection)

    def test_flattenMode_notJsonFile_fails(self):
        """
        When a non-JSON file is passed to the flattening function, an
        exception is raised and the program exits neatly
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.flatten_mode(self.not_json_file, self.this_dir)
        self.assertEqual(context_manager.exception.code, 1)

    def test_flattenMode_corruptJson_fails(self):
        """
        When a corrupt file (doesnt comply with standards) is passed as
        JSON to the flatten function, it should exit neatly
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.flatten_mode(self.bad_json_file, self.this_dir)
        self.assertEqual(context_manager.exception.code, 1)

    def test_flattenMode_notTweetJsonFile_OK(self):
        """
        A non-tweet JSON file should be flattened correctly. Check that
        the operation is performed OK and that the resulting file equals
        one of the example files previously checked for correctness
        """

        flat_file = main.flatten_mode(self.not_tweet_json_file, self.this_dir)
        are_equal = filecmp.cmp(flat_file, self.not_tweet_json_flat_file)
        self.assertTrue(are_equal)

        os.remove(flat_file)

    def test_flattenMode_notTweetJsonCollection_OK(self):
        """
        A non-tweet JSON collection should be flattened correctly. Check
        that this operation is performed correctly.
        Note that date cannot be detected for non-tweet JSONs, so the
        resulting files should be directly placed in the collection
        parent directory
        Check that the resulting collection contains 3 files (the number
        in the original collection)
        """

        flat_collection = main.flatten_mode(
            self.not_tweet_json_collection, self.this_dir)
        len_collection = len(os.listdir(flat_collection))
        self.assertEqual(len_collection, 3)

        shutil.rmtree(flat_collection)

    def test_flattenMode_notJsonDir_fail(self):
        """
        If the input points to any other directory that contains
        non-JSON files, make a neatly fail.
        Check that the program exits with a code 1 due to the exception
        and ensure that no new directory was created for the aborted
        flat collection
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.flatten_mode(self.not_json_dir, self.this_dir)
        self.assertEqual(context_manager.exception.code, 1)

        self.assertFalse(os.path.isdir(os.path.join(self.this_dir, "Flat")))

    def test_flattenMode_alreadyFlatJson_OK(self):
        """
        If the input is an already flat JSON the output should be the
        same as the input
        """

        reflattened_file = main.flatten_mode(
            self.json_flat_file, self.this_dir)

        are_equal = filecmp.cmp(reflattened_file, self.json_flat_file)
        self.assertTrue(are_equal)
        os.remove(reflattened_file)


class Flat2CsvTestCase(unittest.TestCase):

    def setUp(self):
        """Set up some variables and files"""

        self.this_file = os.path.realpath(__file__)
        self.this_dir = os.path.dirname(self.this_file)
        self.input_dir = os.path.join(self.this_dir, "input")

        self.json_file = os.path.join(self.input_dir, "single_tweet.json")
        self.json_flat_file =\
            os.path.join(self.input_dir, "single_tweet_flat.json")
        self.not_json_file = os.path.join(self.input_dir, "text_file.txt")
        self.bad_json_file =\
            os.path.join(self.input_dir, "corrupt_json_file.json")
        self.json_collection = os.path.join(self.input_dir, "ten_tweets")
        self.not_tweet_json_file =\
            os.path.join(self.input_dir, "not_tweet_file.json")
        self.not_tweet_json_flat_file =\
            os.path.join(self.input_dir, "not_tweet_file_flat.json")
        self.not_tweet_json_collection =\
            os.path.join(self.input_dir, "not_tweet_collection")
        self.not_json_dir = os.path.join(self.input_dir, "not_json_dir")
        self.flat_json_collection =\
            os.path.join(self.input_dir, "ten_tweets_flat")
        self.not_fully_flat_collection =\
            os.path.join(self.input_dir, "ten_tweets_flat_one_no_flat")

        main.silent = True

    def test_flat2csv_singleFlatJsonFile_OK(self):
        """
        Given a single flat JSON file as input, the CSV converter
        function should perform as expected and generate a single CSV
        file. Check that file exists.
        """

        csv_file =\
            main.flat2csv_mode(self.json_flat_file, self.this_dir)
        self.assertTrue(os.path.isfile(csv_file))

        os.remove(csv_file)

    def test_flat2csv_singleNotJsonFile_fail(self):
        """
        Given a single non JSON file, the program should make a neat
        fail for not being capable of performing the conversion
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.flat2csv_mode(self.not_json_file, self.this_dir)
        self.assertEqual(context_manager.exception.code, 1)

    def test_flat2csv_singleNotFlatJsonFile_fail(self):
        """
        Given a single JSON file which has not been flattened, the
        program should fail
        NOTE: This could be changed to another way of managing the
        exception in a future
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.flat2csv_mode(self.json_file, self.this_dir)
        self.assertEqual(context_manager.exception.code, 1)

    def test_flat2csv_flatJsonCollection_OK(self):
        """
        Given a collection of flat JSON files, the CSV converter
        function should perform as expected and generate a collection
        of CSV files
        """

        csv_collection =\
            main.flat2csv_mode(self.flat_json_collection, self.this_dir)

        for year_dir in os.listdir(csv_collection):
            year_col = os.path.join(csv_collection, year_dir)

            for month_dir in os.listdir(year_col):
                month_col = os.path.join(year_col, month_dir)

                for day_dir in os.listdir(month_col):
                    day_col = os.path.join(month_col, day_dir)

                    for csv_file in os.listdir(day_col):
                        csv_file_path = os.path.join(day_col, csv_file)

                        self.assertRegex(year_dir, r'\d{4}')
                        self.assertRegex(month_dir, r'\d{2}')
                        self.assertRegex(day_dir, r'\d{2}')
                        self.assertTrue(os.path.isfile(csv_file_path))
                        self.assertEqual(len(os.listdir(day_col)), 10)
                        self.assertRegex(csv_file, r'\d+\.csv$')

        shutil.rmtree(csv_collection)

    def test_flat2csv_notJsonCollection_fail(self):
        """
        Given the path of a directory containing not JSON files, the
        program should exit in an orderly way
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.flat2csv_mode(self.not_json_dir, self.this_dir)
        self.assertEqual(context_manager.exception.code, 1)

    def test_flat2csv_notFlatJsonCollection_fail(self):
        """
        Given a collection of JSON files where at least one of them has
        not been previously flattened, the program should fail
        NOTE: might be changed for another kind of managing
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.flat2csv_mode(self.not_fully_flat_collection, self.this_dir)
        self.assertEqual(context_manager.exception.code, 1)


class SingleJson2CsvTestCase(unittest.TestCase):

    def setUp(self):
        self.this_file = os.path.realpath(__file__)
        self.this_dir = os.path.dirname(self.this_file)
        self.input_dir = os.path.join(self.this_dir, "input")
        self.output_dir = os.path.join(self.this_dir, "output")
        self.project_dir = os.path.dirname(self.this_dir)

        self.json_file = os.path.join(self.input_dir, "single_tweet.json")
        self.json_file2 = os.path.join(self.input_dir, "single_tweet2.json")
        self.csv_file2 =\
            os.path.join(self.input_dir, "single_tweet2.csv")
        self.include_filter =\
            os.path.join(self.input_dir, "default_include_filter.txt")
        self.exclude_filter =\
            os.path.join(self.input_dir, "default_exclude_filter.txt")
        self.corrupt_json =\
            os.path.join(self.input_dir, "corrupt_json_file.json")

    def test_singlejson2csv_singleFileIncludeFilterUpv_OK(self):
        """
        Check that a single JSON file is converted correctly to CSV when
        operating in UPV-compatible mode and with the default UPV filter. That
        is: remove dollar terminations, flatten 1 time the arrays and apply the
        default include filter
        """

        # UPV mode
        main.array_compression = True
        main.compression_depth = 1
        main.remove_dollars = True
        regex_include_filter = main.load_filter(self.include_filter)

        converted_file =\
            main.singlejson2csv_mode(self.json_file2, self.this_dir,
                                     include_filter=regex_include_filter)
        with open(converted_file) as open_converted_file:
            with open(self.csv_file2) as open_csv_file2:
                csv_doc1 = csv.reader(open_converted_file)
                csv_doc2 = csv.reader(open_csv_file2)
                csv_doc1_lines = []
                csv_doc2_lines = []

                for row in csv_doc1:
                    csv_doc1_lines.append(str(row))
                for row in csv_doc2:
                    csv_doc2_lines.append(str(row))

                self.assertEqual(csv_doc1_lines, csv_doc2_lines)
        os.remove(converted_file)

    def test_singlejson2csv_singleFileAlreadyConvertedAnswerNo_exits(self):
        """
        Check that trying to convert a JSON file to CSV when the resulting CSV
        file already exists prompts a message to confirm overwrite Y/N, and
        answering "n" makes the program exit with a 1
        """

        csv_file =\
            main.singlejson2csv_mode(self.json_file, self.this_dir)

        with self.assertRaises(SystemExit) as context_manager:
            with mock.patch('main.input', side_effect=['n']):
                main.singlejson2csv_mode(self.json_file, self.this_dir)

        if os.path.isfile(csv_file):
            os.remove(csv_file)

        self.assertEqual(context_manager.exception.code, 1)

    def test_singlejson2csv_singleFileAlreadyConvertedAnswerYesOverwr_OK(self):
        """
        Check that trying to convert a JSON file to CSV when the resulting CSV
        file already exists prompts a message to confirm overwrite Y/N, and
        answering "y" makes the program overwrite the file. Check that the file
        was overwritten (modification date less than 1 second)
        """

        csv_file =\
            main.singlejson2csv_mode(self.json_file, self.this_dir)

        with mock.patch('main.input', side_effect=['y']):
            csv_file =\
                main.singlejson2csv_mode(self.json_file, self.this_dir)

        now_date = time.time()
        csv_file_date = os.path.getmtime(csv_file)

        self.assertTrue((now_date - csv_file_date) < 1)
        if os.path.isfile(csv_file):
            os.remove(csv_file)

    def test_singlejson2csv_notJsonFile_fails(self):
        """
        Check that trying to apply the JSON to CSV conversion to a not JSON
        file fails in an orderly manner
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.singlejson2csv_mode(self.include_filter, self.this_dir)

        self.assertEqual(context_manager.exception.code, 1)

    def test_singlejson2csv_corruptJsonFile_fails(self):
        """
        Check that trying to apply the JSON to CSV conversion to a file corrupt
        JSON file (with .json extension but that doesn't correctly follow the
        JSON structure) fails in an orderly manner
        """

        with self.assertRaises(SystemExit) as context_manager:
            main.singlejson2csv_mode(self.corrupt_json, self.this_dir)

        self.assertEqual(context_manager.exception.code, 1)

    def test_singlejson2csv_singleFileConvertedKeepFiles_OK(self):
        """
        Check that a single file can be converted correctly and that the
        intermediate files are stored where they belong if keep flag is up
        """

        main.keep_files = True

        csv_file = main.singlejson2csv_mode(self.json_file, self.this_dir)
        flat_file_path = os.path.join(self.this_dir, "single_tweet_flat.json")

        self.assertEqual(os.path.basename(csv_file), "single_tweet.csv")
        self.assertTrue(os.path.isfile(flat_file_path))

        os.remove(csv_file)
        os.remove(os.path.join(self.this_dir, "single_tweet_flat.json"))

        main.keep_files = False

    def test_singlejson2csv_keepFilesDoesNotOverwrite_OK(self):
        """
        Check that when a file is converted to CSV and the keep files flag is
        up, the stored flat file does not overwrite the previous one. Instead,
        the converter adds a number as an ending and keeps both files
        """

        main.keep_files = True

        main.singlejson2csv_mode(self.json_file, self.this_dir)

        with mock.patch('main.input', side_effect=['y']):
            csv_file = main.singlejson2csv_mode(self.json_file, self.this_dir)

        first_file = os.path.join(self.this_dir, "single_tweet_flat.json")
        second_file = os.path.join(self.this_dir, "single_tweet_flat_1.json")

        self.assertTrue(os.path.isfile(first_file))
        self.assertTrue(os.path.isfile(second_file))

        os.remove(first_file)
        os.remove(second_file)
        os.remove(csv_file)

        main.keep_files = False


class JsonCollection2CsvTestCase(unittest.TestCase):

    def setUp(self):
        self.this_file = os.path.realpath(__file__)
        self.this_dir = os.path.dirname(self.this_file)
        self.input_dir = os.path.join(self.this_dir, "input")
        self.project_dir = os.path.dirname(self.this_dir)
        self.output_dir = os.path.join(self.project_dir, "output")

        self.ten_tweets = os.path.join(self.input_dir, "ten_tweets.json")
        self.two_tweets = os.path.join(self.input_dir, "two_tweets.json")
        self.split_ten_tweets = os.path.join(self.input_dir, "ten_tweets")

    def test_jsoncollection2csv_noKeep_OK(self):
        """
        Check that the test makes the full conversion cycle from a collection
        of JSON files to a collection of CSVs correctly. This doesn't keep
        intermediate files, so it is just the CSV collection that needs to be
        cleaned at the end.
        """

        output_dir = main.create_collection_dir(self.ten_tweets, self.this_dir)
        csv_ten_tweets =\
            main.jsoncollection2csv_mode(self.ten_tweets, output_dir)
        i = 0
        for year_dir in os.listdir(csv_ten_tweets):
            year_dir_path = os.path.join(csv_ten_tweets, year_dir)

            for month_dir in os.listdir(year_dir_path):
                month_dir_path = os.path.join(year_dir_path, month_dir)

                for day_dir in os.listdir(month_dir_path):
                    day_dir_path = os.path.join(month_dir_path, day_dir)

                    for json_file in os.listdir(day_dir_path):

                        self.assertRegex(day_dir, r'\d{2}')
                        self.assertRegex(month_dir, r'\d{2}')
                        self.assertRegex(year_dir, r'\d{4}')
                        self.assertRegex(json_file, r'.*.csv')
                        i += 1

        self.assertTrue(i == 10)
        shutil.rmtree(output_dir)

    def test_jsoncollection2csv_keepFiles_OK(self):
        """
        Full conversion cycle from a collection of JSON files to one of CSV
        files, keeping the intermediate results. Check that the 3 collections
        exist (split, flat and csv) and then clean that files. CARE: This will
        fail if the user creates a "ten_tweets" collection in the split or flat
        output folders
        """

        main.keep_files = True
        output_dir = main.create_collection_dir(self.ten_tweets, self.this_dir)
        main.jsoncollection2csv_mode(self.ten_tweets, output_dir)

        self.assertTrue("ten_tweets" in os.listdir(self.this_dir))
        self.assertTrue("Split" in os.listdir(output_dir))
        self.assertTrue("Flat" in os.listdir(output_dir))
        self.assertTrue("CSV" in os.listdir(output_dir))

        shutil.rmtree(output_dir)

        main.keep_files = False


unittest.main()
